//Import PeasyCam https://mrfeinberg.com/peasycam/
import peasy.*;
// import controlP5
import controlP5.*;

PeasyCam cam;
ControlP5 cp5;

//variables for gui interaction
boolean showBox;
boolean chooseRandom;
int randomIndex = 0;
String randomTitle;

//set up the box dimensions as variables so we can use them later in calculations
int boxWidth = 1600;
int boxHeight = 1200;
int boxDepth = 1000;

//manually found start and end year but could also be set in setup function
int startYear = 2006;
int endYear = 2023;
//I'm dealing with time as minutes, so hour the library opens * 60 minutes in the hour = 8*60 
int startTime = 8*60;
int totalDaySpan = 6574;
int totalMinutesSpan = 720;
int totalDaysInYear = 365;

//table which stores CSV data
Table table;
//list of Items which will store each row's data as our Item class
ArrayList<Item> items = new ArrayList<Item>();

//dictionary of all titles in table, and the indices of the data points with those titles
HashMap<String, ArrayList<Integer>> titleMap = new HashMap<String, ArrayList<Integer>>();

void setup() {
  //initalizing our canvas
  size(1000,800,P3D);
  
  // Setup of the camera, the initial distance is 3000 by setting
  cam = new PeasyCam(this, 3000);
  
  // Set the zoom scale of the camera
  cam.setMinimumDistance(200);
  cam.setMaximumDistance(5000);
  
  // Set up the GUI
  cp5 = new ControlP5(this);
  showBox = true;
  chooseRandom = false;
  setGui();
  
  table = loadTable("datawithlove.csv", "header");
  int i=0;
  for (TableRow row : table.rows()) { //for each row in the table
    String title = row.getString("title");
    String coutDate = row.getString("cout_date");
    String coutTime = row.getString("cout_time");
    String itemtype = row.getString("itemtype");

    Item newItem = new Item(title, coutDate, coutTime, itemtype);
    newItem.display(); //displaying each item as it is processed for debugging purposes
    items.add(newItem);
    
    //make dictionary of titles so we can connect titles that have been checked out multiple times
    ArrayList<Integer> prevSeenTitles = titleMap.get(title);
    if (prevSeenTitles == null) {
      prevSeenTitles = new ArrayList<Integer>();
    }
    prevSeenTitles.add(i);
    titleMap.put(title, prevSeenTitles);
    
    i++;
  }
}

void draw() {
  background(75);
  
  //draw the gui
  gui();
  
  //plot each item
  for (int i=0; i< items.size(); i++) {
    //if chooseRandom is on, skip any points that don't have the same title as the chosen data point
    if (chooseRandom) {
      if (!items.get(i).title.equals(randomTitle)) {
        continue; //continue statement skips this iteration of for loop and goes to the next
      }
    }
    
    float yearAxis = getYearAxis(i);
    float timeAxis = getTimeAxis(i);
    float dayAxis = getDayAxis(i);
    
    //check for outliers and skip them
    if (yearAxis < -boxWidth/2 || yearAxis > boxWidth/2 
        || timeAxis < -boxHeight/2 || timeAxis > boxHeight/2 
        || dayAxis < -boxDepth/2 || dayAxis > boxDepth/2) {
      continue;
    }
    
    //draw the points
    pushMatrix();
    translate(yearAxis, timeAxis, dayAxis);
    strokeWeight(3);
    stroke(getColorForItemType(items.get(i).itemtype));
    point(0,0,0);
    
    //title on hover
    float mouseObjectDistance = sq(mouseX-screenX(0,0,0))+sq(mouseY-screenY(0,0,0));
    if (mouseObjectDistance < 10) drawTitle(items.get(i));
    popMatrix();
  }
  
  drawLabels();
  
  noStroke();
  fill(255,40);
  
  // either show filled box or connected lines depending on toggled setting "showBox"
  if (!showBox) {
    drawLines();
    stroke(240);
    strokeWeight(1);
  }
  box(boxWidth, boxHeight, boxDepth);
}

void drawLabels() {
  fill(255);
  textSize(1000);
  pushMatrix();
  textAlign(LEFT, TOP);
  translate(-boxWidth/2, boxHeight/2, boxDepth/2+10);
  scale(0.1);
  text(startYear, 0, 0, 0);
  popMatrix();
  
  pushMatrix();
  textAlign(RIGHT, TOP);
  translate(boxWidth/2, boxHeight/2, boxDepth/2+10);
  scale(0.1);
  text(endYear, 0, 0, 0);
  popMatrix();
  
  pushMatrix();
  textAlign(RIGHT, TOP);
  translate(-boxWidth/2, boxHeight/2-100, boxDepth/2);
  scale(0.1);
  text("8AM", 0, 0, 0);
  popMatrix();
  
  pushMatrix();
  textAlign(RIGHT, BOTTOM);
  translate(-boxWidth/2, -boxHeight/2, boxDepth/2);
  scale(0.1);
  text("8PM", 0, 0, 0);
  popMatrix();
  
  pushMatrix();
  textAlign(RIGHT, TOP);
  translate(boxWidth/2+200, boxHeight/2-100, boxDepth/2);
  scale(0.1);
  text("Jan", 0, 0, 0);
  popMatrix();
  
  pushMatrix();
  textAlign(RIGHT, TOP);
  translate(boxWidth/2+200, boxHeight/2-100, -boxDepth/2);
  scale(0.1);
  text("Dec", 0, 0, 0);
  popMatrix();
}

void drawTitle(Item item) {
  // add a bigger point
  strokeWeight(10);
  point(0,0,0);
  fill(255,255,255);
  // add text label
  textMode(SHAPE);
  // if the distance is close, make the textsize smaller
  if(cam.getDistance()<1700){
    textSize(30);
  }
  // if the distance is far, make the textsize bigger
  else{
    textSize(80);
  }
  text(item.title.toUpperCase(),0,0);
}

//function to connect instances where the same title is checked out with lines
void drawLines() {
  noFill();
  strokeWeight(1);

  for (String title : titleMap.keySet()) { //for every unique title in the table
    //if chooseRandom is on, skip any points that don't have the same title as the chosen data point
    if (chooseRandom) {
      if (!title.equals(randomTitle)) {
        continue; //continue statement skips this iteration of for loop and goes to the next
      }
    }
    
    ArrayList<Integer> itemIndexs = titleMap.get(title); //get all item indices with current title
    if (itemIndexs.size() > 1) { //only titles with multiple checkouts need to be draw as lines
      stroke(getColorForItemType(items.get(itemIndexs.get(0)).itemtype), 90);
      
      beginShape(); //start drawing a connected line
      for (int i = 0; i< itemIndexs.size(); i++) { //for each item with that title
        float yearAxis = getYearAxis(itemIndexs.get(i));
        float timeAxis = getTimeAxis(itemIndexs.get(i));
        float dayAxis = getDayAxis(itemIndexs.get(i));
        
        //check for outliers and skip them
        if (yearAxis < -boxWidth/2 || yearAxis > boxWidth/2 
            || timeAxis < -boxHeight/2 || timeAxis > boxHeight/2 
            || dayAxis < -boxDepth/2 || dayAxis > boxDepth/2) {
          continue;
        }
    
        vertex(yearAxis, timeAxis, dayAxis);
      }
      
      if (itemIndexs.size() == 2) { //there's a bug where lists of two won't draw a line, so in that case add an extra vertex at the first point again
        vertex(getYearAxis(itemIndexs.get(0)), getTimeAxis(itemIndexs.get(0)), getDayAxis(itemIndexs.get(0)));
      }
      
      endShape(); //end connected line
    }
  }
}

void setGui() {
  // Set up the fonts
  PFont pfont = createFont("Arial", 16,true);
  ControlFont font = new ControlFont(pfont, 12);
  float X = 60; //x position for gui elements
  float distance =40; //vertical spacing between gui elements
  int index = 1; //starting at 1 b/c want the first element to be 40 on y-axis
  
  //draw the GUI
  
  //toggle = on/off choice
  cp5.addToggle("showBox")
  .setFont(font)
  .setPosition(X, distance*index);
  index++;
  
  //bang calls a function
  cp5.addBang("turnOnRandomChoice")
  .setFont(font)
  .setPosition(X, distance*index)
  .setLabel("Choose random item");
  index++;
  
  cp5.addTextlabel("label")
  .setFont(font)
  .setText("Press space to turn off random choice")
  .setColorValue(0xffffff00) //can set with hex value, as shown, or RGB or HSV like .setColorValue(color(255,255,0))
  .setPosition(X, index*distance);
  index++;
  
  cp5.setAutoDraw(false);
}

// A function for draw GUI correctly
void gui(){
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}
